/*
 * gapi_app2d.hpp
 *
 * App2D using GAPI (PocketPC game API).
 *
 * (SDL is supposed to support PocketPC, but I couldn't get it to
 * compile on pocketpc-gcc, hence going to use GAPI directly.)
 * 
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef GAPI_APP2D_HPP
#define GAPI_APP2D_HPP

#include "app2d.hpp"

#include <windows.h>
#include "gx.h"
#undef min  // get rid of annoying min and max
#undef max  // macros from windows.h


// AppAlreadyExists:
// This may be thrown by the GapiApp2D ctor. If caught, program should exit silently.
class AppAlreadyExists : public AppError {
public:
    AppAlreadyExists() : AppError("program already running") { }
};


// GapiApp2D class
class GapiApp2D : public App2D {
public:
    GapiApp2D(const TCHAR * window_class_name, const TCHAR * window_title);
    virtual ~GapiApp2D();

    // Full screen displays
    // Note: findDisplayModes and getDisplayMode aren't implemented. Instead, just call
    // openFullScreen with mode_num = 0. (You can get the display width/height after the
    // display is opened.)
    // Note: openFullScreen ignores the "title" argument. Instead you have to specify a
    // Unicode window title in the constructor ...
    virtual int findDisplayModes() { throw AppError("not implemented"); }
    virtual const DisplayMode & getDisplayMode(int n) { throw AppError("not implemented"); }
    virtual void openFullScreen(int mode_num, const char *title);

    // Windowed displays -- Not implemented.
    virtual void openWindowed(int, int, int, int, const char *)
        { throw AppError("not implemented"); }

    // Query functions for displays
    virtual int getDisplayWidth() { return screen_width; }
    virtual int getDisplayHeight() { return screen_height; }
    virtual bool isFullScreen() { return display_open; }
    virtual bool isWindowed() { return false; }
    virtual bool hasFocus() { return isActive(); }

    // Main Loop
    virtual void enterMainLoop();
    virtual void exitMainLoop();
    
    // Access to gapi functions.
    void* gapiBeginDraw() { return GAPI_BeginDraw? GAPI_BeginDraw() : 0; }
    int gapiEndDraw() { return GAPI_EndDraw? GAPI_EndDraw() : 0; }

    // Display properties.
    int getXPitch() const { return props.cbxPitch; }
    int getYPitch() const { return props.cbyPitch; }
    unsigned short rgbToScreen(int r, int g, int b) const
        { return ((r >> 3) << rshift) + ((g >> gclamp) << 5) + (b >> 3); }
    void screenToRgb(unsigned short c, int &r, int &g, int &b) const
    { r = (c & rmask) >> (rshift-3); g = (c & gmask) >> (5-gclamp); b = (c & 0x001f) << 3; }

    // Timer
    virtual int getMilliseconds();
    
protected:
    virtual void resetUpdateIntervals();

public:
    static GapiApp2D * self() { return self_ptr; }
    LRESULT wndProc(HWND, UINT, WPARAM, LPARAM);
    
private:
    static GapiApp2D *self_ptr;
    
    const TCHAR * window_class_name;
    const TCHAR * window_title;
    int screen_width, screen_height;
    bool display_open, render_pending;

    UINT timer;  // current timer object
    HINSTANCE hinstance;   // windows application instance
    HWND hwnd;   // window handle of main window

    // gapi function pointers
    GXDisplayProperties (*GAPI_GetDisplayProperties)();
    GXKeyList (*GAPI_GetDefaultKeys)(int);
    int (*GAPI_OpenDisplay)(HWND, DWORD);
    int (*GAPI_OpenInput)();
    
    int (*GAPI_CloseDisplay)();
    int (*GAPI_CloseInput)();
    
    void* (*GAPI_BeginDraw)();
    int (*GAPI_EndDraw)();
    int (*GAPI_Suspend)();
    int (*GAPI_Resume)();
    int (*GAPI_SetViewport)(DWORD, DWORD, DWORD, DWORD);
    BOOL (*GAPI_IsDisplayDRAMBuffer)();

    // aygshell function: SHFullScreen (to be dynamically linked)
    BOOL (WINAPI *SH_FullScreen)(HWND hwndRequester, DWORD dwState);

    // display properties
    GXDisplayProperties props;
  GXKeyList keylist;
    unsigned short rmask, gmask, bmask;
    int rshift, gclamp, gshift;
};

#endif
