/*
 * enet_controller.hpp
 *
 * This is a simple wrapper over ENet that allows the use of iostreams for the sending
 * and receiving of packets.
 * 
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef ENET_CONTROLLER_HPP
#define ENET_CONTROLLER_HPP

#include "enet/enet.h"

#include "boost/shared_ptr.hpp"
using namespace boost;

#include <exception>
#include <vector>
using namespace std;

class EnetStream;

class EnetController {
public:
    EnetController();
    ~EnetController();

    // Call startServer to start up the server. A UDP port number must be given. Max no of
    // clients and bandwidth limits can be specified. Bandwidth is given in bytes per second
    // (note NOT bits/sec!) and a value of zero means "unlimited".
    // Call getClients to get a list of streams representing the currently connected
    // clients. [TODO-- this could be improved somewhat.... eg interface for allowing/denying
    // client connections, or for booting clients off?]
    void startServer(int port, int maxclients = 32, int bw_incoming = 0,
                     int bw_outgoing = 0);
    void getClients(vector<shared_ptr<EnetStream> > &result);

    // Call connectToServer to start up a client. Returns an EnetStream representing the
    // connection to the server, or 0 if connection failed.
    // Address can be given either as a hostname, or as a dotted IP address. Bandwidth limits
    // are as above. The timeout is in ms -- this routine blocks until connection succeeds
    // or the timeout expires. (Could do something asynchronous in future I guess.)
    shared_ptr<EnetStream> connectToServer(const char *address, int port, int timeout,
                                           int bw_incoming = 0, int bw_outgoing = 0);

    // Call shutDown to close down all connections. This will block for up to "timeout"
    // (in ms) before forcefully resetting connections.
    void shutDown(int timeout);

    // processMessages: Clear out the ENet message queue. Usually this is called automatically
    // as needed by stream operations, but can be called manually (say, once per game loop),
    // which may be needed if no network I/O is being done. Also can be useful for blocking,
    // eg while waiting for clients to connect.
    void processMessages(int timeout = 0) { processMessages(timeout, true); }

    bool isConnected() const { return server || client; }

    string getIP() const;
    
private:
    friend class EnetStreambuf;
    void processMessages(int timeout, bool ac); // called by EnetStreambuf
    
private:
    ENetHost *server;  // set if we are a server
    struct PeerData {
        ENetPeer *peer;
        shared_ptr<EnetStream> stream;
    };
    vector<PeerData> server_peers;
    
    ENetHost *client;  // set if we are a client
    PeerData client_peer;
};


//
// Error class:
//
class EnetError : public exception {
public:
    explicit EnetError(const char *err) : err_string(err) { }
    virtual const char * what() const throw() { return err_string; }
private:
    const char * err_string;
};


#endif
