/*
 * enet_stream.hpp
 *
 * EnetStreambuf: A streambuf for ENet networking.
 * EnetStream: An iostream using an EnetStreambuf.
 *
 * Notes:
 * 
 * For input, we store up all the received packets so far. Check in_avail() to
 * see how much data is available. Reading beyond the end of the buffer causes an
 * EOF error (we never block).
 *
 * For output, written data is buffered up indefinitely -- the buffer must be
 * explicitly flushed before data is sent. When this is done, we write out all the
 * stored data as a single packet.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef ENET_STREAM_HPP
#define ENET_STREAM_HPP

#include "enet/enet.h"

#include <deque>
#include <iostream>
#include <streambuf>
#include <vector>
using namespace std;

class EnetController;

class EnetStreambuf : public streambuf {
public:
	// Note: ctor is private; EnetStreambufs can only be created by an
	// EnetController (via an EnetStream).
	~EnetStreambuf();
	
protected:
	// A minimal streambuf implementation only needs to override
	// overflow, underflow and sync:
	virtual int_type overflow(int_type meta = traits_type::eof());
	virtual int_type underflow();
	virtual int sync();

	// Here we also provide showmanyc:
	virtual streamsize showmanyc();

private:
	friend class EnetStream;
	EnetStreambuf(EnetController *n, ENetPeer *p) : net(n), peer(p) { }
	void close();
	bool isOpen() const { return net; }
	void receivePacket(ENetPacket* p);

private:
	void clearInputBuffer();
	streamsize updateInputBuffer();
	
private:
	// References to the EnetController and the ENetPeer:
	EnetController *net;
	ENetPeer *peer;
	
	// Input buffer:
	// We just store up pointers to the received packets that ENet supplies us with.
	// (We destroy the packets when the buffer underflows, or when dtor is called.)
	deque<ENetPacket*> input_buffer;

	// Output buffer:
	// This can grow indefinitely in size. The streambuf must be flushed explicitly
	// before anything actually gets written.
	vector<char> output_buffer;
};

class EnetStream : public iostream {
public:
	// This will return false if the connection has been lost. (If a connection is closed
	// then I/O operations will fail with an EOF error.)
	bool isOpen() const { return buf.isOpen(); }
	
private:
	friend class EnetController;
	EnetStream(EnetController *net, ENetPeer *p) : iostream(0), buf(net, p) { init(&buf); }
	void close() { buf.close(); }
	void receivePacket(ENetPacket *p) { buf.receivePacket(p); }
	
private:
	EnetStreambuf buf;
};

#endif
