/*
 * gfx_resizer_compose.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "gfx_resizer_compose.hpp"
#include "round.hpp"

#include <cmath>
using namespace std;

float GfxResizerCompose::intermedScaleFactor(float ideal_scale_factor) const
{
    if (!right) return 1;  // "right" is assumed to be the identity in this case.
    
    float intermed_down, intermed_up;
    right->roundScaleFactor(ideal_scale_factor, intermed_down, intermed_up);

    // we use intermed_down or intermed_up, whichever is closer
    const float diff_down = fabs(intermed_down - ideal_scale_factor);
    const float diff_up = fabs(intermed_up - ideal_scale_factor);
    
    return (diff_up < diff_down) ? intermed_up : intermed_down;
}

void GfxResizerCompose::roundScaleFactor(float ideal_scale_factor, float &final_down, float &final_up) const
{
    if (lock_to_int) {
        // If lock_to_int is set then first we round ideal_scale_factor to an int, THEN we round that to
        // whatever the scalers want. So we don't guarantee an integer scaling; we merely get as close as we
        // can to that, given what the scalers want to do.
        // Note also: We always round down rather than up.
        ideal_scale_factor = int(ideal_scale_factor);
    }
    
    // Work out what the intermediate will scale to.
    const float intermed = intermedScaleFactor(ideal_scale_factor);
    
    // We now have a graphic of size intermed times the original
    // This must be scaled by ideal_scale_factor / intermed to get to the required final size.
    left->roundScaleFactor(ideal_scale_factor / intermed, final_down, final_up);

    // correct for the fact that we did the intermediate scaling!
    final_down *= intermed;
    final_up *= intermed;
}

shared_array<RGB> GfxResizerCompose::resize(shared_array<RGB> original, int old_width, int old_height,
                                            int new_width, int new_height) const
{
    const float ideal_scale_factor = float(new_height)/float(old_height);

    const float intermed_scale_factor = intermedScaleFactor(ideal_scale_factor);
    const int intermed_width = Round(old_width * intermed_scale_factor);
    const int intermed_height = Round(old_height * intermed_scale_factor);

    shared_array<RGB> intermed = right ? right->resize(original, old_width, old_height, intermed_width, intermed_height) : original;
    shared_array<RGB> final = left->resize(intermed, intermed_width, intermed_height, new_width, new_height);
    return final;
}
