/*
 * sdl_graphic_cache.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "display.hpp"
#include "gfx_resizer.hpp"
#include "graphic_transform.hpp"
#include "sdl_graphic.hpp"
#include "sdl_graphic_cache.hpp"

SdlGraphicCache g_sdl_graphic_cache;

SdlGraphicCache::SdlGraphicCache()
    : npix(0), display(0)
{
}

SdlGraphicCache::~SdlGraphicCache()
{
    clear();
}

void SdlGraphicCache::clear()
{
    for (GraphicMap::iterator it = graphic_map.begin(); it != graphic_map.end(); ++it) {
        delete it->second.new_graphic;
    }
    graphic_map.clear();
    recent_list.clear();
    npix = 0;
}

const Graphic * SdlGraphicCache::createGraphic(const GraphicKey &key)
{
    // NOTE: This routine should be called with either a resize or a colour-change,
    // but not both at the same time.
    
    if (key.cc.empty()) {
        // Resize only.
        ASSERT(key.new_width != GetWidth(*key.original) || key.new_height != GetHeight(*key.original));
        
        // Call the resizing routine.
        return CreateResizedGraphic(*display->getGfxResizer(), *key.original, key.new_width, key.new_height);
        
    } else {
        
        // Colour change only.
        ASSERT(key.new_width == GetWidth(*key.original) && key.new_height == GetHeight(*key.original));

        // Call the colour-changing routine.
        return CreateGraphicWithCC(*key.original, key.cc);
    }
}

const Graphic * SdlGraphicCache::getGraphic(const GraphicKey &key)
{
    GraphicMap::iterator it = graphic_map.find(key);
    if (it == graphic_map.end()) {
        // Not in cache; need to create a new graphic
        const Graphic * new_graphic = createGraphic(key);
        
        // Add to the RecentList (at the front)
        recent_list.push_front(key);

        // Add to the main cache.
        GraphicData gd;
        gd.new_graphic = new_graphic;
        gd.list_position = recent_list.begin();
        graphic_map.insert(make_pair(key, gd));

        // Increase npix
        npix += key.new_width * key.new_height;

        // Delete old graphics if necessary.
        deleteOld();

        // Return the new graphic.
        return new_graphic;
        
    } else {
        // It is in the cache. Bump it to the front of the MRU list.
        recent_list.splice(recent_list.begin(), recent_list, it->second.list_position);
        return it->second.new_graphic;
    }
}
    
const Graphic * SdlGraphicCache::getGraphicWithCC(const Graphic &original, const ColourChange &cc)
{
    // Short cut:
    if (cc.empty()) return &original;

    // Look it up in cache (creating if necessary):
    GraphicKey key;
    key.original = &original;
    key.cc = cc;
    key.new_width = GetWidth(original);
    key.new_height = GetHeight(original);
    return getGraphic(key);
}
    
const Graphic * SdlGraphicCache::getResizedGraphic(const Graphic &original, int new_width, int new_height)
{
    // Short cut:
    if (new_width == original.getWidth() && new_height == original.getHeight()) return &original;

    // Look it up in cache (creating if necessary):
    GraphicKey key;
    key.original = &original;
    key.cc = ColourChange();
    key.new_width = new_width;
    key.new_height = new_height;
    return getGraphic(key);
}

const Graphic * SdlGraphicCache::getResizedGraphicWithCC(const Graphic &original, int new_width, int new_height, const ColourChange &cc)
{
    // This is just a composition of the previous two operations.
    const Graphic * gfx_cc = getGraphicWithCC(original, cc);
    return getResizedGraphic(*gfx_cc, new_width, new_height);
}

bool SdlGraphicCache::tooBig() const
{
    // NOTE: minimum_number should be at least 2, otherwise getResizedGraphicWithCC
    // may fail (because its intermediate "gfx_cc" may get deleted before we can use it!).
    
    const int minimum_number = 5;
    const int maximum_number = 1000;
    const int maximum_pixels = 1000000;

    const int current_number = graphic_map.size();

    if (current_number <= minimum_number) return false;
    if (current_number > maximum_number) return true;
    if (npix > maximum_pixels) return true;
    return false;
}

void SdlGraphicCache::deleteOld()
{
    // Delete old graphic(s) if the cache is too big.
    while (tooBig()) {
        ASSERT(!recent_list.empty());

        // Find the least recently accessed graphic in the GraphicMap. We are going to remove it.
        GraphicMap::iterator it = graphic_map.find(recent_list.back());
        ASSERT(it != graphic_map.end());
        
        // Reduce npix accordingly
        npix -= it->first.new_width * it->first.new_height;

        // Delete graphic, and remove it from GraphicMap and RecentList
        delete it->second.new_graphic;
        graphic_map.erase(it);
        recent_list.pop_back();
    }
}
